import type { ErrorResponse, SuccessResponse, FilterKeys, MediaType, PathsWithMethod, ResponseObjectMap, OperationRequestBodyContent, HasRequiredKeys } from "openapi-typescript-helpers";
/** options for each client instance */
interface ClientOptions extends Omit<RequestInit, "headers"> {
    /** set the common root URL for all API requests */
    baseUrl?: string;
    /** custom fetch (defaults to globalThis.fetch) */
    fetch?: typeof fetch;
    /** global querySerializer */
    querySerializer?: QuerySerializer<unknown>;
    /** global bodySerializer */
    bodySerializer?: BodySerializer<unknown>;
    headers?: HeadersOptions;
}
export type HeadersOptions = HeadersInit | Record<string, string | number | boolean | null | undefined>;
export type QuerySerializer<T> = (query: T extends {
    parameters: any;
} ? NonNullable<T["parameters"]["query"]> : Record<string, unknown>) => string;
export type BodySerializer<T> = (body: OperationRequestBodyContent<T>) => any;
export type ParseAs = "json" | "text" | "blob" | "arrayBuffer" | "stream";
export interface DefaultParamsOption {
    params?: {
        query?: Record<string, unknown>;
    };
}
export type ParamsOption<T> = T extends {
    parameters: any;
} ? HasRequiredKeys<T["parameters"]> extends never ? {
    params?: T["parameters"];
} : {
    params: T["parameters"];
} : DefaultParamsOption;
export type RequestBodyOption<T> = OperationRequestBodyContent<T> extends never ? {
    body?: never;
} : undefined extends OperationRequestBodyContent<T> ? {
    body?: OperationRequestBodyContent<T>;
} : {
    body: OperationRequestBodyContent<T>;
};
export type FetchOptions<T> = RequestOptions<T> & Omit<RequestInit, "body">;
export type FetchResponse<T> = {
    data: FilterKeys<SuccessResponse<ResponseObjectMap<T>>, MediaType>;
    error?: never;
    response: Response;
} | {
    data?: never;
    error: FilterKeys<ErrorResponse<ResponseObjectMap<T>>, MediaType>;
    response: Response;
};
export type RequestOptions<T> = ParamsOption<T> & RequestBodyOption<T> & {
    querySerializer?: QuerySerializer<T>;
    bodySerializer?: BodySerializer<T>;
    parseAs?: ParseAs;
    fetch?: ClientOptions["fetch"];
};
export default function createClient<Paths extends {}>(clientOptions?: ClientOptions): {
    /** Call a GET endpoint */
    GET<P extends PathsWithMethod<Paths, "get">>(url: P, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P], "get">>> extends never ? [(FetchOptions<FilterKeys<Paths[P], "get">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P], "get">>]): Promise<FetchResponse<"get" extends infer T ? T extends "get" ? T extends keyof Paths[P] ? Paths[P][T] : unknown : never : never>>;
    /** Call a PUT endpoint */
    PUT<P_1 extends PathsWithMethod<Paths, "put">>(url: P_1, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P_1], "put">>> extends never ? [(FetchOptions<FilterKeys<Paths[P_1], "put">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P_1], "put">>]): Promise<FetchResponse<"put" extends infer T_1 ? T_1 extends "put" ? T_1 extends keyof Paths[P_1] ? Paths[P_1][T_1] : unknown : never : never>>;
    /** Call a POST endpoint */
    POST<P_2 extends PathsWithMethod<Paths, "post">>(url: P_2, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P_2], "post">>> extends never ? [(FetchOptions<FilterKeys<Paths[P_2], "post">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P_2], "post">>]): Promise<FetchResponse<"post" extends infer T_2 ? T_2 extends "post" ? T_2 extends keyof Paths[P_2] ? Paths[P_2][T_2] : unknown : never : never>>;
    /** Call a DELETE endpoint */
    DELETE<P_3 extends PathsWithMethod<Paths, "delete">>(url: P_3, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P_3], "delete">>> extends never ? [(FetchOptions<FilterKeys<Paths[P_3], "delete">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P_3], "delete">>]): Promise<FetchResponse<"delete" extends infer T_3 ? T_3 extends "delete" ? T_3 extends keyof Paths[P_3] ? Paths[P_3][T_3] : unknown : never : never>>;
    /** Call a OPTIONS endpoint */
    OPTIONS<P_4 extends PathsWithMethod<Paths, "options">>(url: P_4, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P_4], "options">>> extends never ? [(FetchOptions<FilterKeys<Paths[P_4], "options">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P_4], "options">>]): Promise<FetchResponse<"options" extends infer T_4 ? T_4 extends "options" ? T_4 extends keyof Paths[P_4] ? Paths[P_4][T_4] : unknown : never : never>>;
    /** Call a HEAD endpoint */
    HEAD<P_5 extends PathsWithMethod<Paths, "head">>(url: P_5, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P_5], "head">>> extends never ? [(FetchOptions<FilterKeys<Paths[P_5], "head">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P_5], "head">>]): Promise<FetchResponse<"head" extends infer T_5 ? T_5 extends "head" ? T_5 extends keyof Paths[P_5] ? Paths[P_5][T_5] : unknown : never : never>>;
    /** Call a PATCH endpoint */
    PATCH<P_6 extends PathsWithMethod<Paths, "patch">>(url: P_6, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P_6], "patch">>> extends never ? [(FetchOptions<FilterKeys<Paths[P_6], "patch">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P_6], "patch">>]): Promise<FetchResponse<"patch" extends infer T_6 ? T_6 extends "patch" ? T_6 extends keyof Paths[P_6] ? Paths[P_6][T_6] : unknown : never : never>>;
    /** Call a TRACE endpoint */
    TRACE<P_7 extends PathsWithMethod<Paths, "trace">>(url: P_7, ...init: HasRequiredKeys<FetchOptions<FilterKeys<Paths[P_7], "trace">>> extends never ? [(FetchOptions<FilterKeys<Paths[P_7], "trace">> | undefined)?] : [FetchOptions<FilterKeys<Paths[P_7], "trace">>]): Promise<FetchResponse<"trace" extends infer T_7 ? T_7 extends "trace" ? T_7 extends keyof Paths[P_7] ? Paths[P_7][T_7] : unknown : never : never>>;
};
/** serialize query params to string */
export declare function defaultQuerySerializer<T = unknown>(q: T): string;
/** serialize different query param schema types to a string */
export declare function defaultQueryParamSerializer<T = unknown>(key: string[], value: T): string | undefined;
/** serialize body object to string */
export declare function defaultBodySerializer<T>(body: T): string;
/** Construct URL string from baseUrl and handle path and query params */
export declare function createFinalURL<O>(pathname: string, options: {
    baseUrl: string;
    params: {
        query?: Record<string, unknown>;
        path?: Record<string, unknown>;
    };
    querySerializer: QuerySerializer<O>;
}): string;
/** merge headers a and b, with b taking priority */
export declare function mergeHeaders(...allHeaders: (HeadersOptions | undefined)[]): Headers;
export {};
